/*
 * The base of source is ZDOC [ Ver.5.0.1b3 ].
 *
 * This program is distributed under the terms of the GPL v2.0 or later
 * Download the GNU Public License (GPL) from www.gnu.org
 * 
 */
// These are the basic routines that read and write records to the database.
// DOC files of version 1, 2, and 3 can be handled.
// Based on the version of the currently loaded file, the appropriate
// decompression (if any) is done after reading, and the appropriate compression
// is done before writing.

#define DISP_BITS 11
#define COUNT_BITS 3

#include <PalmOS.h>
#include "pvm1_c.h"
#include "pvm1_g.h"

// Forward declarations of functions for sectioning purposes
static UInt16 compress(Buffer * b) EXTRA_SECTION_ONE;
static Boolean compress2(Buffer * b) EXTRA_SECTION_ONE;
static void DecompressDoc(Buffer * b) EXTRA_SECTION_ONE;
static void issue(Buffer * b, unsigned char src) EXTRA_SECTION_ONE;
static unsigned char *memfind2(unsigned char *t, unsigned char *m, Int16 m_len) EXTRA_SECTION_ONE;
static UInt16 uncompress(Buffer * b) EXTRA_SECTION_ONE;

// This was formerly DOC_Text.cook, but the elaborate
// structure was no longer needed.
char CookBuffer[BUFSIZE];

// Replace the given buffer with an uncompressed version of itself.
// The supplied buffer must have version 2 compression
// input: b - pointer to buffer with data compressed via version 2
// output: b - pointer to a buffer with uncompressed result
// Note: data is not returned in the same buffer in which
// it was supplied!
static UInt16 uncompress(Buffer * b)
{
   unsigned char *in_buf;
   unsigned char *out_buf;
   Int16 j = 0, i = 0;

   out_buf = CookBuffer;
   in_buf = b->buf;

   for (i = j = 0; i < b->len;) {
      UInt16 c = in_buf[i++];

// ##ry pyrite.org palm DOC spec suggests that 9 is a legal count
// but programs such as DocReader consider 8 to be the max count
//      if (c > 0 && c <= 9) {
      if (c > 0 && c < 9) {
	 while (c--) {
	    out_buf[j++] = in_buf[i++];
	 }
      } else if (c < 0x80) {
	 out_buf[j++] = c;
      } else if (c >= 0xc0) {
	 out_buf[j++] = ' ';
	 out_buf[j++] = c ^ 0x80;
      } else {
	 Int16 di, n;
	 c = (c << 8) + in_buf[i++];
	 di = (c & 0x3fff) >> COUNT_BITS;
	 n = (c & ((1 << COUNT_BITS) - 1)) + 3;
	 while (n--) {
	    out_buf[j] = out_buf[j - di];
	    j++;
	 }
      }

      // Quit decompressing if we might overflow the buffer
      if (j > MAXTEXTSIZE) {
         j = MAXTEXTSIZE;
         break;
      }
   }

   b->buf = out_buf;
   b->len = j;
   return j;
}

// Decompress a buffer
// Slightly different from the original doc_uncompress in that this
// does not copy the buffer to CookBuffer if not needed
// Version 2 compression will be handled by calling uncompress()
// Version 3 encryption will be handled in-line
// Version 1 (no compression) or any other version is passed verbatim
// input: b - pointer to buffer to uncompress
// output b = pointer to uncompressed result
// The buffer will be uncompressed according to the
// method corresponding to DOC_Values.version
// Note that output buffer will be re-used the next time either
// compression or uncompression is called, so the caller should
// copy the result right away if it needs to stay around
static void DecompressDoc(Buffer * b)
{
   // decompress if needed
   if (DOC_Values.version == 2) {
      uncompress(b);
   }

   // decrypt if needed
   if (DOC_Values.version == 3) {
     UInt16 a, i;
     unsigned char *buf;

     buf = b->buf;
     i = 0;
      for (a = 0; a < b->len; a++) {
	 buf[a] = (256 - buf[a]) - key[i];
	 i++;
	 if (i >= KEYLEN) {
	    i = 0;
	 }
      }
   }

   // Important: for use in a Field, append trailing null
   *(b->buf + b->len) = '\0';
}

// For performance measurement, set to 1
#define TEST 0

// read record specified by index into supplied buffer */
// inputs: dbR - handle to database
//         index - record number to load
//         b -> buf - must point to an array of char
// output: b->buf - filled with result
//         b->len - updated with length
// return value: true - successful read
//               false - unsuccessful read, probably attempt to read beyond EOF
// Note: b->buf assumed to be large enough.  For a DOC file,
// records are <= 4096 bytes, so b->buf had better be 4097
// to accomodate the trailing null.
// Why copy the information to the supplied buffer,
// rather than simply locking the handle supplied by
// DmQueryRecord?  It's trouble to keep track of the
// handle later for unlocking, and could result in
// a handle leak.  A clever programmer could do it, but
// I need to have everything working in my favor.  - ry
Boolean ReadRec(DmOpenRef dbR, UInt16 index, Buffer *b)
{

   MemPtr recPtr;
   MemHandle recHand;

   UInt32 len;

#if TEST
   UInt32 start_time;
#endif

   // Show where we are
   StrPrintF(OutString, "\253%u", index); // "<<RecNumber"
   StatusMsg(OutString, false);

#if TEST
   start_time = TimGetTicks();
#endif

   // a record beyond the known ones
   // can happen with badly-formed files
   if (index > DOC_Values.count) {
      // StatusMsg("Err 037", true);  no warning, could be common
      b->buf[0] = '\0';
      b -> len = 0;
      return false;
   }

   recHand = DmQueryRecord(dbR, index);
   // Requested record <= DOC_Values.count, but doesn't exist
   if (recHand == NULL) {
      StatusMsg("Prog Err 03", true);
      b->buf[0] = '\0';
      b -> len = 0;
      return false;
   }

   recPtr = MemHandleLock(recHand);
   len = MemPtrSize(recPtr);
   if (len > MAXTEXTSIZE) {
      StatusMsg("Rec too big!", true);
      len = MAXTEXTSIZE;  // safeguard just in case record too large
   }

   MemMove(b->buf, recPtr, len);

   MemHandleUnlock(recHand);
   // DmReleaseRecord(dbR, index, false);  Not needed, we only did a DmQueryRecord
   // MemHandleFree(recHand);  DO NOT free the memory, it deletes the record!

   b -> len = len;

   DecompressDoc(b);

#if TEST
   AlertNum("RR time diff:", TimGetTicks() - start_time);
#endif

   return true;
}

#if 0
// This appears to be a defunct string-search routine
// Search for string m of length m_len within
// string t of length t_len.
// Return is a pointer to the place in t where m was found,
// or if not found, returns 0 (null)
static unsigned char *memfind(unsigned char *t, Int16 t_len, unsigned char *m,
			      Int16 m_len)
{
   Int16 i;

   for (i = t_len - m_len + 1; i > 0; i--, t++)
      if (t[0] == m[0] && MemCmp(t, m, m_len) == 0)
	 return t;
   return 0;
}
#endif

// This looks like a more sophisticated string search,
// tuned for better performance.  It is fast because
// we avoid indexing on the search string in common cases.
// It appears a lot of thought has gone into this; if
// you seek to improve upon it, take care to performance
// test your code by defining TEST below.
// Search for string m of length m_len within
// string t.  Return a pointer to the place in t where m
// was found.
// It is assumed that m is a part of t, so eventually, m
// will be found, even if the pointer returned is to itself.
static unsigned char *memfind2(unsigned char *t, unsigned char *m, Int16 m_len)
{
   unsigned char c0, c1, c2, c3;

   if (m_len == 1) {
      c0 = m[0];
      while (true) {
	 if (t[0] == c0) {
	    return t;
	 }
	 t++;
      }
   } else if (m_len == 2) {
      c0 = m[0];
      c1 = m[1];
      while (true) {
	 if (t[0] == c0 && t[1] == c1) {
	    return t;
	 }
	 t++;
      }
   } else if (m_len == 3) {
      c0 = m[0];
      c1 = m[1];
      c2 = m[2];
      while (true) {
	 if (t[0] == c0 && t[1] == c1 && t[2] == c2) {
	    return t;
	 }
	 t++;
      }
   } else if (m_len == 4) {
      c0 = m[0];
      c1 = m[1];
      c2 = m[2];
      c3 = m[3];
      while (true) {
	 if (t[0] == c0 && t[1] == c1 && t[2] == c2 && t[3] == c3) {
	    return t;
	 }
	 t++;
      }
   } else {
      c0 = m[0];
      c1 = m[1];
      c2 = m[2];
      c3 = m[3];
      while (true) {
	 if (t[0] == c0 && t[1] == c1 && t[2] == c2 && t[3] == c3) {
	    if (MemCmp(t, m, m_len) == 0) {
	       return t;
	    }
	 }
	 t++;
      }
   }
   return 0;
}

#if 0
// Here is a failed attempt that was not faster; in fact,
// it took twice as long.  To know why, you need to understand
// the Motorola architecture.
// Search for string m of length m_len within
// string t.  Return a pointer to the place in t where m
// was found.
// It is assumed that m is a part of t, so eventually, m
// will be found, even if the pointer returned is to itself.
static unsigned char *memfind3(unsigned char *tp, unsigned char *mp, Int16 m_len)
{
 register unsigned char *m0;  // the start of the string sought
 register unsigned char *t0;  // the place in t being considered for matching m
 register unsigned char *m;  // fast copies of passed parameters
 register unsigned char *t;

 m0 = m = mp;
 t0 = t = tp;

 // pointers t and m will continue to advance as long as the characters
 // match
 while (true) {
    if ((m-m0) >= m_len)  // found it!
       return t0;

    if (*t != *m) {
       // Doesn't match, move along in t, and start over in m
       t = ++t0;
       m = m0;
    } else {
       // Matches so far, advance.
       t++;
       m++;
    }
 }
}
#endif

// This global is used during compression
// True if the character before the one presently
// being processed was a space (' ')
static Boolean bSpace;

// This is part of the type 2 compression, and handles
// the space compression case, as well as unusual
// characters that need the quote verbatim prefix.
// Verbatim characters are quoted one a a time.
// These are consolidated in a later step.
// The flag bSpace is set to true if the previous character
// before src was a space.
static void issue(Buffer * b, unsigned char src)
{
   Int16 iDest = b->len;
   unsigned char *dest = b->buf;

   if (bSpace) {
      bSpace = false;
      if (src >= 0x40 && src <= 0x7f) {
	 dest[iDest++] = src ^ 0x80;
	 b->len = iDest;
	 return;
      }
      dest[iDest++] = ' ';
   } else {
      if (src == ' ') {
	 bSpace = true;
	 b->len = iDest;
	 return;
      }
   }
   // ##ry changed from 8 to 9, to adhere to DOC compression spec
   // found on pyrite.org
   // Note that the value 0 is automatically ignored (edited out)
   // if it is ever found.  It would mess up the doc, anyway. 
   // if ((src > 0 && src <= 9) || src >= 0x80) {
   if ((src >= 1 && src <= 8) || src >= 0x80) {
      dest[iDest++] = 1;
   }
   dest[iDest++] = src;
   b->len = iDest;
   return;
}

// Replace the given buffer with a compressed version of itself.
// Input: b - buffer structure, pointing to input data
// Output: b - buffer structure, now points to output data
// WARNING! The output data is not simply put into the same
// buffer where the input data was supplied!
static UInt16 compress(Buffer * b)
{
   Int16 i;

   unsigned char *pBuffer;  // Original start of input buffer
   unsigned char *pHit;     // Where the test string was found
   unsigned char *pPrevHit; // Where the search starts for test string
   unsigned char *pTestHead; // The start of the test string
   unsigned char *pTestTail; // (One char after) the end of test string
   unsigned char *pEnd;      // (one char after) the end of input buffer

   UInt16 dist, compound, k;

#if TEST
   UInt32 start_time;

   start_time = TimGetSeconds();
#endif

   bSpace = false;

   // Set up all the input pointers based on passed parameter
   // because b->buf is about to be switched
   pHit = pPrevHit = pTestHead = pBuffer = b->buf;
   pTestTail = pTestHead + 1;
   pEnd = b->buf + b->len;

   // Switch in the output buffer
   b->buf = CookBuffer;
   b->len = 0;

   while (pTestHead != pEnd) {
      // String replace can only back-reference a certain
      // number of bits.  Make sure we're not searching
      // too far back for reference string.
      if (pTestHead - pPrevHit > ((1 << DISP_BITS) - 1))
	 pPrevHit = pTestHead - ((1 << DISP_BITS) - 1);

      // Here's the core: find the test string in range
      // pPrevHit .. test string itself
      pHit = memfind2(pPrevHit, pTestHead, pTestTail - pTestHead);

      if (pHit == pTestHead
	  || pTestTail - pTestHead > (1 << COUNT_BITS) + 2
	  || pTestTail == pEnd) {
         // a condition occurred whereby we can no longer
         // keep testing for a longer string.  either
         // 1) a character mismatch, or
         // 2) string would exceed maximum replacement size of 10, or
         // 3) hit the end of input
	 if (pTestTail - pTestHead < 4) {
            // If matching test string is too short, there
            // would be no gain by using the string-replace
            // to compress it.  Simply send the character out.
	    issue(b, pTestHead[0]);
	    pTestHead++;
	 } else {
            // The test string is suitable for the string-replace.
            // First, output any pending space.  The replacement
            // string may contain or even start with a space
            // but if bSpace was set, then the pending space
            // was not part of the test string.
	    if (bSpace) {
	       b->buf[b->len++] = ' ';
	       bSpace = false;
	    }
            // Encode the string-replace, and pass to output buffer
	    dist = pTestHead - pPrevHit;
	    compound = (dist << COUNT_BITS) + pTestTail - pTestHead - 4;
	    b->buf[b->len++] = 0x80 + (compound >> 8);
	    b->buf[b->len++] = compound & 0xff;
	    pTestHead = pTestTail - 1;
	 }
	 pPrevHit = pBuffer;
      } else {
         // The string matched, On the next pass, we will
         // increase the test string size by 1.  But we already
         // know that we need not search before where this
         // string was found, so save time by starting the
         // search there.
	 pPrevHit = pHit;
      }

      // Lengthen the test string by 1
      if (pTestTail != pEnd)
	 pTestTail++;
   } // end of while loop


   // Emit a final pending space, if any
   if (bSpace)
      b->buf[b->len++] = ' ';

   // The verbatim characters were quoted one at a time
   // by the issue() routine above.  This consolidates
   // consecutive quotes into a single quote, up to
   // the limit of 8.
   for (i = k = 0; i < b->len; i++, k++) {
      b->buf[k] = b->buf[i];
      if (b->buf[k] >= 0x80 && b->buf[k] < 0xc0)
	 b->buf[++k] = b->buf[++i];
      else if (b->buf[k] == 1) {
	 b->buf[k + 1] = b->buf[i + 1];

         // ##ry changed from 8 to 9, to adhere to DOC compression spec
         // found on pyrite.org
	 //while (i + 2 < b->len && b->buf[i + 2] == 1 && b->buf[k] < 9) {
         while (i + 2 < b->len && b->buf[i + 2] == 1 && b->buf[k] < 8) {
	    b->buf[k]++;
	    b->buf[k + b->buf[k]] = b->buf[i + 3];
	    i += 2;
	 }
	 k += b->buf[k];
	 i++;
      }
   }

   b->len = k;
#if TEST
   AlertNum("Time diff:", TimGetSeconds() - start_time);
#endif
   return k;
}

// Replace the given buffer with a compressed version of itself.
// Input: b - buffer structure, pointing to input data
// Output: b - buffer structure, now points to output data
// Return: true, if succeeded
//         false, if compressed output wouldn't fit in buffer
// This experimental version does NOT use string-replace.
// The hope is to be able to handle compressed files, but
// with reasonable speed.  There is no expectation that
// any sizeable compression will actually occur.  In fact,
// this scheme may be unusable if there are lots of characters
// not between 0x0a and 0x7f, for example, accented characters.
// WARNING! The output data is not simply put into the same
// buffer where the input data was supplied!
// The source buffer must have space for a terminating null
// after all its data.
static Boolean compress2(Buffer * b)
{
 unsigned char *pSource;    // Points to char under consideration in source
 unsigned char src;         // Source character being examined
 unsigned char *pDest;      // Points to place to be written in destination
 unsigned char *pDestEnd;   // Should never attempt to write here or beyond
 unsigned char *pDestCount; // The count byte for quoting verbatim
 char QuoteCount;  // How many strange bytes are quoted

 pSource = b -> buf;
 *(pSource + b->len) = '\0';  // Terminate the input buffer

 pDest = CookBuffer;
 pDestEnd = pDest + MAXTEXTSIZE;

 do {
    src = *pSource;

    // Space is the most common character, treat it first
    if (src == ' ') {
       src = *++pSource;
       if (src >= 0x40 && src <= 0x7f) {
          // suitable for space compression
	  *pDest++ = src ^ 0x80;
          pSource++;
          continue;  // go back to top, maybe next char is a space, too
       } else {
          // not suitable, just eject a space
          *pDest++ = ' ';
       }
    }

   // In English at least, almost all the other characters will
   // fall into this case of pass-through
   if (0x09 <= src && src <= 0x7f) {
      *pDest++ = src;
      pSource ++;
   } else {

      if (src==0) break;  // quit if we hit end of text

      // Quote verbatim characters - up to 9
      pDestCount = pDest;  // remember where to put the count later
      pDest++;
      if (pDest >= pDestEnd) break;  // overflow - no space to write

      QuoteCount = 0;
      // pyrite.org spec may disagree with defacto standard
//      while ((src >= 0x80 || (0 < src && src <= 9)) && QuoteCount < 9) {
      while ((src >= 0x80 || (0 < src && src <= 8)) && QuoteCount < 8) {
          if (pDest >= pDestEnd) break;  // overflow - quit and update count
          *pDest++ = src;
          QuoteCount++;
          src = *++pSource;
      }
      *pDestCount = QuoteCount;  // fill in the count
   }

   if (pDest >= pDestEnd) break;  // overflow - no space to write

 } while (src != 0);

 // if we broke out of the while loop for reasons other than
 // hitting the end of the source buffer
 if (src != 0) {
    // AlertNum("QCompress dropped chars:", b->buf + b->len - pSource);
    StatusMsg("Qcomp drop", true);
 }

 // Return the output buffer
 b -> buf = CookBuffer;
 b -> len = (UInt32) pDest - (UInt32) (b -> buf);

 return (src==0);  // Return true if successful
}

// Write the DOC header in the first record
// This was evolved from WriteValues() in the
// original.  As input, this is passed the local ID of
// the database, not an open reference.  If the header record does not
// exist, (for example, when creating a new file), one is created.
void WriteHeader(LocalID dbID)
{
   DmOpenRef dbW;
   MemHandle recHand;
   MemPtr recPtr;
   Err err;
   UInt16 at;

   dbW = DmOpenDatabase(0, dbID, dmModeReadWrite);

   // if first record doesn't exist, create one
   if (DmNumRecords(dbW) == 0) {
      at = 0;
      recHand = DmNewRecord(dbW, &at, sizeof DOC_Values);
   } else {
      recHand = DmGetRecord(dbW, 0);
   }

   ErrFatalDisplayIf(recHand == 0, "Sorry, fatal error (50) writing header. Probably out of memory.");
   recPtr = MemHandleLock(recHand);
   ErrFatalDisplayIf(recPtr == 0, "Sorry, fatal error (51) writing header. Probably out of memory.");
   err = DmWrite(recPtr, 0, &DOC_Values, sizeof DOC_Values);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (52) writing header. Probably out of memory.");
   //   MemHandleUnlock(recHand);
   err = MemPtrUnlock(recPtr);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (53) writing header. Probably out of memory.");
   err = DmReleaseRecord(dbW, 0, true);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (55) writing header. Probably out of memory.");

   DmCloseDatabase(dbW);
}

// Take an uncompressed buffer, format it if necessary,
// then write to disk.
// This is done by creating a new record at the index
// specified, then deleting the original record.  This allows
// the new record to be a different size from the old.
// To do a replace, set the input variable "new" to false.
// The database is opened for writing for a minimum time.
// Inputs: dbID - Local ID of Database to write
//         index - index of record to write
//         new - if true, don't delete the old record at index
void WriteRec(LocalID dbID, Boolean newrec, UInt16 index, Buffer *b)
{
   DmOpenRef dbW;
   MemHandle recHand; // = NULL;
   MemPtr recPtr; // = NULL;
   Err err;
   char * SaveBuf;  // use to temporarily save buffer in
   UInt16 SaveLen;  // case a restart is needed

#if TEST
   UInt32 start_time;
#endif

   // Show where we are
   StrPrintF(OutString, "\273%u", index); // ">>RecNumber"
   StatusMsg(OutString, false);

#if TEST
   start_time = TimGetTicks();
#endif

   // Palm does not permit generating zero-length data in records.
   // If one is requested, we will make one of length 1.  We will
   // choose a character that will neither expand nor compress with
   // the compression scheme, yet will be unlikely to match any
   // text in the field.  If I program properly, this safety net
   // should never be encountered.
   if (b->len == 0) {
      AlertNum("Programming error: Attempt to write zero-length record @", index);
      b->len = 1;
      b->buf[0] = 0x7f;
   }

   // Compress if necessary
   if (DOC_Values.version == 2) {
      SaveBuf = b->buf;
      SaveLen = b->len;
      // (rare) If user demands full compression, or
      // (also rare) quick compress algorithm is unable
      // to generate a record smaller than 4096 bytes
      if (!compress2(b)) {
         StatusMsg("std cpress", false);
         b->buf = SaveBuf;
         b->len = SaveLen;
         compress(b);
      }
   }

   // Encrypt if necessary
   if (DOC_Values.version == 3) {
      UInt16 a, i;
      unsigned char *buf;

      buf = b->buf;
      i = 0;
      for (a = 0; a < b->len; a++) {
	 buf[a] = 256 - (buf[a] + key[i]);
	 i++;
	 if (i >= KEYLEN) {
	    i = 0;
	 }
      }
   }

   dbW = DmOpenDatabase(0, dbID, dmModeReadWrite);

   recHand = DmNewRecord(dbW, &index, b->len);
   if (DmGetLastErr())
      AlertNum("WriteRec error: ", DmGetLastErr());
   ErrFatalDisplayIf(recHand == 0, "Sorry, fatal error (60) writing record. Probably out of memory.");
   recPtr = MemHandleLock(recHand);
   ErrFatalDisplayIf(recPtr == 0, "Sorry, fatal error (61) writing record. Probably out of memory.");
   err = DmWrite(recPtr, 0, b->buf, b->len);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (62) writing record. Probably out of memory.");
   //   MemHandleUnlock(recHand);
   err = MemPtrUnlock(recPtr);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (63) writing record. Probably out of memory.");
   err = DmReleaseRecord(dbW, index, true);
   ErrFatalDisplayIf(err != 0, "Sorry, fatal error (64) writing record. Probably out of memory.");

   if (!newrec) {
      DmRemoveRecord(dbW, index + 1);
   } else {
      DOC_Values.count++;  // new record added
   }

   DmCloseDatabase(dbW);

#if TEST
   AlertNum("WR time diff:", TimGetTicks() - start_time);
#endif

}

// Delete the record at index
// input: index - index of record to delete
// First checks to see if the record exists.  If it does not,
// no delete is attempted, as this generates a fatal error
// on some releases of PalmOS.
void DelRec(LocalID dbID, UInt16 index)
{
   DmOpenRef dbW;

   dbW = DmOpenDatabase(0, dbID, dmModeReadWrite);
   if (DmQueryRecord(dbW, index)) {
      DmRemoveRecord(dbW, index);
      DOC_Values.count--;
   }
   DmCloseDatabase(dbW);
}